# -*- coding: utf-8 -*-
import pandas as pd
from . .queries import *
from . .queries import neg_response
from . .figures import *
from IPython.display import clear_output
from statsmodels.sandbox.regression.predstd import wls_prediction_std
from . .curve_fitting import *

def lin_response(log_file_path):
    valid_responses = ['yes', 'y', 'true', 't', 'no', 'n', 'false', 'f']
    while True:
        response = input("\nAssign a linearity correction? (Y/N)\n").lower()
        if response in valid_responses:
            append_to_log(log_file_path, "Linearity application application: "+str(response))
            return response
        else:
            print("\nInvalid response. Try again.\n")

def process_linearity_correction(samp, drift, lin_std, user_choice, correction_log, folder_path, fig_path, isotope, user_linearity_conditions, log_file_path):
    ex = pd.DataFrame()
    dD_id = 'drift_corrected_dD' if correction_log.loc["Drift","samples"] == 1 else "dD"
    # Normalize data
    norm=pd.DataFrame()
    mean_isotope_dict={}
    for i in ["C20","C28"]:
        mask = lin_std.chain==i
        temp = lin_std.loc[mask, ['area','chain',dD_id]].copy()
        mean_isotope_dict[i] = temp[dD_id].mean()
        #temp[user_choice] -= temp[user_choice].mean() # For gls regression
        temp[user_choice] -= temp[user_choice].min()-1
        norm = pd.concat([norm, temp])
    response = lin_response(log_file_path)
    if neg_response(response):
        print("\nSkipping linearity correction.\n")
        return samp, correction_log, lin_std, drift
    else:
        if user_linearity_conditions:
            while True:
                chain_corr_val = input("\nEnter peak area cutoff value for C_{20} and C_{28} linearity stds\n")
                try:
                    chain_corr_vals = [float(num) for num in chain_corr_val.split()]  # Split the input string into a list
                    if len(chain_corr_vals) == 1:
                        correction_log.loc["Linearity"] = chain_corr_vals[0]  # Save the single value
                        break  # Break the loop since the condition is met
                    else:
                        print("Please enter one number.\n")
                except ValueError:
                    print("Invalid input. Please enter a numerical value.\n")
        else:
            chain_corr_val = 0
            correction_log.loc["Linearity"] = chain_corr_val
    chain_corr_val = float(chain_corr_val)
    verify_lin_plot(norm, fig_path, dD_id,log_file_path, cutoff_line=chain_corr_val, isotope=isotope)

    user_input = input("\nDoes this look correct? (Y/N)\n").lower()
    if pos_response(user_input):
        append_to_log(log_file_path, "Minimum peak area to derive linearity correction: "+str(chain_corr_val))
        lin_std, drift, samp, excluded_drift, excluded_lin_std, excluded_samp  = linearity_correction(drift, samp, lin_std, norm, chain_corr_val, dD_id, folder_path, fig_path, log_file_path=log_file_path)
        append_to_log(log_file_path, "Number of drift standards excluded because of below threshold area: "+str(len(excluded_drift)))
        append_to_log(log_file_path, "Number of linearity standards excluded because of below threshold area: "+str(len(excluded_lin_std)))
        append_to_log(log_file_path, "Number of samples excluded because of below threshold area: "+str(len(excluded_samp)))
        # Save excluded samples if any
        if not excluded_drift.empty or not excluded_lin_std.empty or not excluded_samp.empty:
            subfolder = create_subfolder(folder_path, "Excluded_Data")
            if not excluded_drift.empty:
                excluded_drift.to_csv(os.path.join(subfolder, 'Drift_standards_excluded_peak_area.csv'), index=False)
            if not excluded_lin_std.empty:
                excluded_lin_std.to_csv(os.path.join(subfolder, 'Linearity_standards_excluded_peak_area.csv'), index=False)
            if not excluded_samp.empty:
                excluded_samp.to_csv(os.path.join(subfolder, 'Samples_excluded_by_peak_area.csv'), index=False)
            #log_excluded_samples(subfolder, excluded_drift, excluded_lin_std, excluded_samp)

        clear_output(wait=True)
        return drift, correction_log, lin_std, samp

    elif neg_response(user_input):
        print("\nSkipping linearity correction.\n")
        time.sleep(0)  # Wait for 1 second
        clear_output(wait=True)
        return drift, correction_log, lin_std, samp

    else:
        time.sleep(0)
        clear_output(wait=True)
        print("\nInvalid response. Try again.\n")

# def linearity_correction(drift, samp, lin_std, lin_norm, area_cutoff, dD_id, folder_path, fig_path, log_file_path, fig=False):
#     area_cutoff = float(area_cutoff)
#     corrected_drift = pd.DataFrame()  # Initialize an empty DataFrame to store corrected unknown
#
#     # Filter linearity standards and unknown based on area cutoff
#     mask                   = lin_norm['area'] >= area_cutoff
#     filtered_lin_norm      = lin_norm[mask] # normalized linearity standards
#     filtered_lin_std       = lin_std[mask]  # original linearity standards
#     filtered_drift         = drift[drift['area'] >= area_cutoff]
#
#     # Store and remove excluded unknown
#     excluded_drift = drift[drift['area']<area_cutoff]
#     excluded_lin_std = lin_std[lin_std['area']<area_cutoff]
#     excluded_samp = samp[samp['area'] < area_cutoff]
#
#     # Store and remove excluded unknown
#     xdata = filtered_lin_norm['area'].to_numpy()
#     ydata = filtered_lin_norm[dD_id].to_numpy()
#     best_model, popt, sse, pcov = fit_and_select_best(xdata, ydata)
#     pred_error = prediction_std(best_model, xdata, popt, pcov)
#     # Generate fitted predictions from the chosen model
#     if best_model == "exponential":
#         y_fit = exp_func(xdata, *popt)
#     else:
#         y_fit = log_func(xdata, *popt)
#
#     # Total Sum of Squares
#     tss = np.sum((ydata - ydata.mean()) ** 2)
#     if tss == 0:
#         r_squared = 1.0  # Degenerate case if data is all the same
#     else:
#         r_squared = 1 - (sse / tss)
#
#     print(f"Chosen Model: {best_model}")
#     print(f"Parameters: {popt}")
#     print(f"SSE: {sse:.3f}, R²: {r_squared:.3f}")
#
#     lin_top_sort = filtered_lin_norm.sort_values(by='area', ascending=False)
#     top_count = max(int(len(lin_top_sort) * 0.2), 1)  # ensure at least 1 point
#     lin_top_qt_area = lin_top_sort.head(top_count)
#     lin_reference = lin_top_qt_area[dD_id].mean()  # average δD of top 20%
#
#     # Combine new fit (y_fit) with the old reference approach
#     lin_cor = y_fit - lin_reference + filtered_lin_norm[dD_id]
#     filtered_lin_std['linearity_corrected_dD'] = lin_cor
#     filtered_lin_std['linearity_error'] = pred_error  # or calculate a custom uncertainty
#
#     # Apply to drift
#     if best_model == "exponential":
#         drift_est = exp_func(np.array(filtered_drift.area), *popt)
#     else:
#         drift_est = log_func(np.array(filtered_drift.area), *popt)
#     pred_error = prediction_std(best_model, np.array(filtered_drift.area), popt, pcov)
#     drift_cor = drift_est-lin_reference
#     filtered_drift['linearity_corrected_dD'] = drift_cor[~np.isnan(drift_cor)]#[~drift_cor.isna()]
#     filtered_drift['linearity_error'] = pred_error
#
#     # Apply to samples
#     filtered_samp = samp[samp['area'] >= area_cutoff]
#     if best_model == "exponential":
#         samp_est = exp_func(np.array(filtered_samp.area), *popt)
#     else:
#         samp_est = log_func(np.array(filtered_samp.area), *popt)
#     pred_error = prediction_std(best_model, np.array(filtered_samp.area), popt, pcov)
#
#     samp_cor = samp_est-lin_reference
#     samp_cor = samp_cor+filtered_samp[dD_id]
#     filtered_samp['linearity_corrected_dD'] = samp_cor[~samp_cor.isna()]
#     filtered_samp['linearity_error'] = pred_error
#
#     return filtered_lin_std, filtered_drift, filtered_samp, excluded_drift, excluded_lin_std, excluded_samp

def linearity_correction(drift, samp, lin_std, lin_norm, area_cutoff, dD_id, folder_path, fig_path, log_file_path, fig=False):
    area_cutoff = float(area_cutoff)
    corrected_drift = pd.DataFrame()  # Initialize an empty DataFrame to store corrected unknown

    # Filter linearity standards and unknown based on area cutoff
    mask              = lin_norm['area'] >= area_cutoff
    filtered_lin_norm = lin_norm[mask].copy()  # copy to avoid SettingWithCopy warning
    filtered_lin_std  = lin_std[mask].copy()
    filtered_drift    = drift[drift['area'] >= area_cutoff].copy()

    # Store excluded data
    excluded_drift   = drift[drift['area'] < area_cutoff]
    excluded_lin_std = lin_std[lin_std['area'] < area_cutoff]
    excluded_samp    = samp[samp['area'] < area_cutoff]

    # ----------------- Fit Models and Select Best -----------------
    xdata = filtered_lin_norm['area'].to_numpy()
    ydata = filtered_lin_norm[dD_id].to_numpy()
    best_model, popt, sse, pcov = fit_and_select_best(xdata, ydata)
    pred_error = prediction_std(best_model, xdata, popt, pcov)
    print(popt)
    # Generate fitted predictions from the chosen model
    if best_model == "linear":
        y_fit = linear_func(xdata, *popt)
    else:  # best_model == "log"
        y_fit = log_func(xdata, *popt)

    # Total Sum of Squares and R²
    tss = np.sum((ydata - ydata.mean()) ** 2)
    if tss == 0:
        r_squared = 1.0
    else:
        r_squared = 1 - (sse / tss)

    print(f"Chosen Model: {best_model}")
    print(f"Parameters: {popt}")
    print(f"SSE: {sse:.3f}, R²: {r_squared:.3f}")

    # Determine reference value from the top 20% (by area) of the linearity standards
    lin_top_sort = filtered_lin_norm.sort_values(by='area', ascending=False)
    top_count = max(int(len(lin_top_sort) * 0.2), 1)
    lin_top_qt_area = lin_top_sort.head(top_count)
    lin_reference = lin_top_qt_area[dD_id].mean()  # average δD of top 20%

    # Combine new fit with the old reference approach for linearity standards
    lin_cor = y_fit - lin_reference + filtered_lin_norm[dD_id]
    filtered_lin_std['linearity_corrected_dD'] = lin_cor
    filtered_lin_std['linearity_error'] = pred_error  # or calculate a custom uncertainty

    # Apply correction to drift
    if best_model == "linear":
        drift_est = linear_func(np.array(filtered_drift.area), *popt)
    else:
        drift_est = log_func(np.array(filtered_drift.area), *popt)
    pred_error_drift = prediction_std(best_model, np.array(filtered_drift.area), popt, pcov)
    drift_cor = drift_est - lin_reference
    filtered_drift['linearity_corrected_dD'] = drift_cor[~np.isnan(drift_cor)]
    filtered_drift['linearity_error'] = pred_error_drift

    # Apply correction to samples
    filtered_samp = samp[samp['area'] >= area_cutoff].copy()
    if best_model == "linear":
        samp_est = linear_func(np.array(filtered_samp.area), *popt)
    else:
        samp_est = log_func(np.array(filtered_samp.area), *popt)
    pred_error_samp = prediction_std(best_model, np.array(filtered_samp.area), popt, pcov)
    samp_cor = samp_est - lin_reference + filtered_samp[dD_id]
    filtered_samp['linearity_corrected_dD'] = samp_cor[~samp_cor.isna()]
    filtered_samp['linearity_error'] = pred_error_samp

    return filtered_lin_std, filtered_drift, filtered_samp, excluded_drift, excluded_lin_std, excluded_samp




























import pandas as pd
from utils.base_functions import *

def calculate_methanol_dD(unknown, isotope, log_file_path):
    """
    Calculation of dD and dC of methanol from PAME analysis
    """
    # Split samples and pames into seperate dataframes
    pame_unknown =  unknown[unknown['chain']=="Phthalic acid"]
    samples =  unknown[unknown['chain']!="Phthalic acid"]
    pame_uk = pame_unknown.copy()
    pame_uk.loc[:,'PAME_methanol_dD'] = pd.NA
    phthalic_original = float(q_original_phthalic_value())
    pame_uk.loc[pame_uk['chain']=="Phthalic acid", 'PAME_methanol_dD'] = (10 * pame_uk.loc[pame_uk['chain']=="Phthalic acid", 'VSMOW_dD'] - 4 * phthalic_original) / 6
    append_to_log(log_file_path, "Calculated methanol values from PAMEs: "+str(pame_uk.loc[pame_uk['chain']=="Phthalic acid", 'PAME_methanol_dD']))
    pame_uk = pame_uk.dropna(axis=1, how='all')
    return samples, pame_uk


from .base_functions import append_to_log
import statsmodels.api as sm
import numpy as np

def wls_regression(x, y, log_file_path): #should be wls_regression - changing it here so that wls can be used without renaming ever instance
    """
    Weighted Least Squares Regression Model Function. Weights are calcualted as the inverse of the amplitude. 

    Parameters:
    - x: Independent variable(s), should be a 1D or 2D array-like structure.
    - y: Dependent variable, should be a 1D array-like structure.

    Returns:
    - slope: Slope coefficient(s) of the model.
    - intercept: Intercept of the model.
    - r_squared: Coefficient of determination.
    - p_value: p-value associated with the slope(s).
    - std_err: Standard error of the slope coefficient(s).
    - results: The results object from the regression model.
    """
    
    append_to_log(log_file_path, "Applied weighted least squares regression function")
    x_with_const = sm.add_constant(x)
    weights = 1 / (np.abs(x) + 1)  # Adding 1 to avoid division by zero, assuming x is your independent variable array
    model = sm.WLS(y, x_with_const, weights=weights)
    results = model.fit()
    
    intercept, slope = results.params
    r_squared = results.rsquared
    p_value = results.pvalues.iloc[1]  # Assuming the slope is the second parameter
    std_err = results.bse.iloc[1]  # Standard error for the slope

    return slope, intercept, r_squared, p_value, std_err, results


import tkinter as tk
import pandas as pd

# Default values
default_stds = {
    "drift_stds": ["C20", "C28"],
    "linearity_stds": ["C18", "C24"]    
}

vsmow_dD = {
    "type":           ["drift", "linearity", "linearity", "drift"],
    "id":             ["C18", "C20", "C28", "C24"],
    "isotope value":  [-206.2, -166.7, -89.28190994, -179.3],
    "std":            [1.7, 0.3, 1.062744893, 1.7],
    "n":              [5, 3, 924, 5],
    "VSMOW accuracy check": ["False", "False", "False", "True"]}

vsmow_dC = {
        "type":           ["drift", "linearity", "drift"],
        "id":             ["C18", "C20", "C24"],
        "isotope value" : [-23.24, -30.68, -26.57],
        "std":            [0.01, 0.02, 0.02],
        "n":              [5,3,5],
        "VSMOW accuracy check": ["False", "False", "True"]}
        
class Editor:
    def __init__(self, alt_stds, isotope):
        self.alt_stds = alt_stds
        self.isotope = isotope
        self.standards_df = pd.DataFrame()  # Initialize the DataFrame
        self.root = tk.Tk()
        self.root.title("Edit Known Values")
        self.create_widgets()
    
    def create_widgets(self):
        if not self.alt_stds:
            # If alt_stds is False, no need to create widgets
            return
        # Select the appropriate data dictionary based on isotope type
        if self.isotope == "dD":
            self.data_dict = vsmow_dD
            isotope_header = "δD VSMOW"
        elif self.isotope == "dC":
            self.data_dict = vsmow_dC
            isotope_header = "δ13C VSMOW"
        else:
            print(f"Unsupported isotope type: {self.isotope}")
            self.root.destroy()
            return
        
        # Define columns
        columns = ["type", "chain length", isotope_header, "standard deviation", "n", "VSMOW accuracy check"]
        
        # Create column headers
        for j, column in enumerate(columns):
            header = tk.Label(self.root, text=column)
            header.grid(row=0, column=j, padx=5, pady=5)  # Place headers in the first row
    
        self.entries = []  # List to store entry widgets
    
        num_rows = len(self.data_dict["id"])  # Determine number of rows based on data dictionary
    
        for i in range(num_rows):  # Create rows based on the data dictionary
            row_entries = []
            for j in range(len(columns)):
                entry = tk.Entry(self.root)
                entry.grid(row=i+1, column=j, padx=5, pady=5)
    
                # Pre-fill values based on row and column
                if j == 0:  # Type
                    entry.insert(0, self.data_dict["type"][i])
                elif j == 1:  # Chain length
                    entry.insert(0, self.data_dict["id"][i])
                elif j == 2:  # δD or δ13C
                    entry.insert(0, self.data_dict["isotope value"][i])
                elif j == 3:  # std
                    entry.insert(0, self.data_dict["std"][i])
                elif j == 4:  # n
                    entry.insert(0, self.data_dict["n"][i])
                elif j == 5:  # accuracy check
                    entry.insert(0, self.data_dict["VSMOW accuracy check"][i])
            
                row_entries.append(entry)
    
            self.entries.append(row_entries)  # Store row entries for access later

        # Create a button to save the values
        save_button = tk.Button(self.root, text="Set Values", command=self.save_and_close)
        save_button.grid(row=5, columnspan=len(columns), pady=10)  # Place save button below the table

    def save_and_close(self):
        """Save values and close the editor window."""
        data = {
            "type": [],
            "chain length": [],
            "isotope value": [],  # Will be renamed based on isotope type
            "std": [],
            "n": [],
            "VSMOW accuracy check": []
        }
    
        # Determine the correct column name based on isotope type
        if self.isotope == "dD":
            isotope_column = "δD"
        elif self.isotope == "dC":
            isotope_column = "δ13C"
        else:
            print("Error: ", f"Unsupported isotope type: {self.isotope}")
            return
    
        try:
            for i in range(len(self.data_dict["id"])):  # Loop through the rows
                data["type"].append(self.entries[i][0].get())
                data["chain length"].append(self.entries[i][1].get())
                data["isotope value"].append(float(self.entries[i][2].get()))
                data["std"].append(float(self.entries[i][3].get()))
                data["n"].append(int(self.entries[i][4].get()))
                
                # Handle "accuracy check" as a boolean
                accuracy_value = self.entries[i][5].get()
                if accuracy_value.lower() == "true":
                    data["VSMOW accuracy check"].append(True)
                elif accuracy_value.lower() == "false":
                    data["VSMOW accuracy check"].append(False)
                else:
                    raise ValueError(f"Invalid accuracy check value: {accuracy_value}")
        except ValueError as e:
            print("Input Error: ", f"Invalid input: {e}")
            return
    
        # Rename the isotope column in the DataFrame
        self.standards_df = pd.DataFrame(data).rename(columns={isotope_column: self.isotope})
        self.root.quit()  # Exit the main loop
        self.root.destroy()  # Destroy the window
    
    def run(self):
        self.root.mainloop()  # Start the Tkinter event loop
        return self.standards_df  # Return the DataFrame after the loop exits

def open_editor(alt_stds, isotope):
    if alt_stds:
        editor = Editor(alt_stds, isotope)
        df = editor.run()
        return df
    else:
        # Build the DataFrame directly from the appropriate dictionary
        if isotope == "dD":
            data = {
                "type": vsmow_dD["type"],
                "chain length": vsmow_dD["id"],
                "δD": vsmow_dD["isotope value"],
                "std": vsmow_dD["std"],
                "n": vsmow_dD["n"],
                "VSMOW accuracy check": [val.lower() == "true" for val in vsmow_dD["VSMOW accuracy check"]]
            }
            return pd.DataFrame(data)
        elif isotope == "dC":
            data = {
                "type": vsmow_dC["type"],
                "chain length": vsmow_dC["id"],
                "δ13C": vsmow_dC["isotope value"],
                "std": vsmow_dC["std"],
                "n": vsmow_dC["n"],
                "VSMOW accuracy check": [val.lower() == "true" for val in vsmow_dC["VSMOW accuracy check"]]
            }
            return pd.DataFrame(data)
        else:
            print(f"Unsupported isotope type: {isotope}")
            return pd.DataFrame()



import numpy as np
import os
import pandas as pd
from .figures import total_dD_correction_plot

def output_results(raw_unknown, unknown, sd, unknown_pame, folder_path, fig_path, res_path, isotope, pame):
    # Define column name mappings for output CSV files
    column_name_mapping = {
        'Identifier 1': 'Sample Name',
        'chain': 'Chain Length',
        'area_mean': 'Mean Area',
        'dD_mean': f'Average raw {isotope}',
        'dD_std': f'{isotope} Std Dev',
        'dD_count': 'Number of Replicates',
        'drift_corrected_dD_mean': f'Drift Corrected {isotope}',
        'drift_error_mean': 'Drift Error',
        'linearity_corrected_dD_mean': f'Linearity Corrected {isotope}',
        'linearity_error_mean': 'Linearity Error',
        'VSMOW_dD_mean': f'VSMOW Corrected {isotope}',
        'VSMOW_error_mean': 'VSMOW Error',
        'methanol_dD_mean': f'Methanol Corrected {isotope}',
        'methanol_error_mean': f'Methanol Error',
        'PAME_methanol_dD_mean': f'PAME Methanol Calculated {isotope}',
        'PAME_methanol_dD_std': f'PAME Methanol Calculated Error {isotope}',
        'replicate_dD_sem': f'Mean replicate std {isotope}',
        'total_uncertainty': 'Total Uncertainty'
    }
    
    # Rename columns in unknown dataframe
    unknown_renamed = unknown.rename(columns=column_name_mapping)
    if pame:
        unknown_pame_renamed = unknown_pame.rename(columns=column_name_mapping)
    if isotope=='dD':
        unknown_renamed.insert(len(unknown_renamed.columns) - 1, 'Corrected '+str(isotope), unknown_renamed['Methanol Corrected '+str(isotope)])
        if pame:
            unknown_pame_renamed.insert(len(unknown_pame_renamed.columns) - 1, 'Corrected Methanol '+str(isotope), unknown_pame_renamed['PAME Methanol Calculated '+str(isotope)])
    else:
        unknown_renamed.reset_index(drop=True, inplace=True)
        unknown_renamed.insert(len(unknown_renamed.columns) - 1, 'Corrected '+str(isotope), unknown_renamed['VSMOW Corrected dC'])
        if pame:
            unknown_pame_renamed.reset_index(drop=True, inplace=True)
            unknown_pame_renamed.insert(len(unknown_pame_renamed.columns) - 1, 'Corrected '+str(isotope), unknown_pame_renamed['VSMOW Corrected dC'])
    # Save unknown dataframe to CSV
    unknown_renamed.reset_index(drop=True, inplace=True)
    unknown_renamed.to_csv(os.path.join(res_path,'Results - sample mean.csv'), index=False)
    if pame:
        unknown_pame_renamed.reset_index(drop=True, inplace=True)
        unknown_pame_renamed.to_csv(os.path.join(res_path,'PAME Results - sample mean.csv'), index=False)
    # Select columns for standards dataframe
    columns_to_select = [
        "Date", "Time", "Identifier 1", "chain", "Rt", "area", "Area 2", "Area 3",
        "Ampl  2", "Ampl  3", "BGD 2", 'BGD 3', "time_rel", "dD", "drift_corrected_dD",
        "drift_error", "linearity_corrected_dD", "linearity_error", "VSMOW_dD", "vsmow_error",
        "total_uncertainty"
    ]

    # Select only existing columns from sd dataframe
    existing_columns = [col for col in columns_to_select if col in sd.columns]
    standards_selected = sd[existing_columns].copy()

    # Filter and categorize standards
    lin_std_temp = standards_selected[standards_selected["Identifier 1"].str.contains("C20|C28")].copy()
    lin_std_temp['Standard Type'] = "Linearity"
    
    drift_std_temp = standards_selected[standards_selected["Identifier 1"].str.contains("C218|C24")].copy()
    drift_std_temp['Standard Type'] = "Drift"
    
    standards_categorized = pd.concat([lin_std_temp, drift_std_temp])

    # Rename columns in standards dataframe
    column_rename_map = {
        "Identifier 1": "Standard ID", "chain": "Component", "Rt": "Retention time",
        "area": "Peak area", "Area 2": "Peak area 2", "Area 3": "Peak area 3",
        "Ampl  2": "Amplitude 2", "Ampl  3": "Amplitude 3", "BGD 2": "Background 2",
        "BGD 3": "Background 3", "time_rel": "Time relative", "dD": f"Raw {isotope}",
        "drift_corrected_dD": f"Drift corrected {isotope}", "drift_error": f"Drift {isotope} error",
        "linearity_corrected_dD": f"Linearity {isotope}", "linearity_error": f"Linearity {isotope} error",
        "VSMOW_dD": f"VSMOW corrected {isotope}", "vsmow_error": "VSMOW error",
        "methanol_dD": f"Methanol corrected {isotope}", "methanol_error": f"Methanol error {isotope}",
        "total_uncertainty": "Total uncertainty"
    }
    standards_categorized = standards_categorized.rename(columns=column_rename_map)
    standards_categorized.insert(len(standards_categorized.columns) - 1, 'Corrected '+str(isotope), standards_categorized[f"VSMOW corrected {isotope}"])
    # Save standards dataframe to CSV
    standards_categorized.to_csv(os.path.join(res_path, 'Results - standards.csv'), index=False)

    # Rename columns in raw_unknown dataframe
    raw_unknown_renamed = raw_unknown.rename(columns=column_rename_map)

    # Drop 'total_error' column if exists
    if 'total_error' in raw_unknown.columns:
        raw_unknown_renamed = raw_unknown_renamed.drop('total_error', axis=1)

    # Save raw_unknown dataframe to CSV
    raw_unknown_renamed.to_csv(os.path.join(res_path, 'Results - sample replicates.csv'), index=False)
    total_dD_correction_plot(raw_unknown_renamed, unknown_renamed, folder_path, fig_path, isotope)
    print("\nCorrections complete :)")

def mean_values_with_uncertainty(data, iso, sample_name_header="Identifier 1", chain_header="chain"):
    # Group by sample name and chain length
    grouped = data.groupby([sample_name_header, chain_header])

    # Start building the aggregation dictionary based on iso
    agg_dict = {
        'area': ['mean'],
        'drift_corrected_dD': 'mean',
        'drift_error': 'mean',
        'linearity_corrected_dD': 'mean',
        'linearity_error': 'mean',
        'VSMOW_dD': ['mean', 'std'],
        'VSMOW_error': 'mean'
    }

    if iso == "dC":
        agg_dict.update({
            'dC': ['mean', 'std', 'count'],
        })
    else:
        agg_dict.update({
            'dD': ['mean', 'std', 'count'],
        })

        # Check if 'methanol_dD' exists in the dataframe
        if 'methanol_dD' in data.columns:
            agg_dict.update({
                'methanol_dD': ['mean', 'std'],
                'methanol_error': 'mean'
            })

        # Check if 'pame_methanol_dD' exists in the dataframe
        if 'PAME_methanol_dD' in data.columns:
            agg_dict.update({
                'PAME_methanol_dD': ['mean', 'std']
            })

    # Apply the aggregation
    stats = grouped.agg(agg_dict).reset_index()

    # Flatten MultiIndex columns
    stats.columns = ['_'.join(col).strip('_') for col in stats.columns.values]
    
    # Calculate SEM for methanol dD (std / sqrt(count)) if iso is 'dD' and methanol_dD exists
    if iso == 'dD':
        if 'methanol_dD_std' in stats.columns and 'dD_count' in stats.columns:
            stats['replicate_dD_sem'] = stats['methanol_dD_std'] / np.sqrt(stats['dD_count'])
        elif 'PAME_methanol_dD_std' in stats.columns and 'dD_count' in stats.columns:
            stats['replicate_dD_sem'] = stats['PAME_methanol_dD_std'] / np.sqrt(stats['dD_count'])
    else:
        stats['replicate_dC_sem'] = stats['VSMOW_error_mean'] / np.sqrt(stats['dC_count'])

    # Calculate total uncertainty
    if iso == 'dD':
        if 'methanol_error_mean' in stats.columns:
            error_columns = ['drift_error_mean', 'linearity_error_mean', 'VSMOW_error_mean', 'methanol_error_mean', 'replicate_dD_sem']
        elif 'PAME_methanol_dD_std' in stats.columns:
            error_columns = ['drift_error_mean', 'linearity_error_mean', 'VSMOW_error_mean', 'replicate_dD_sem']
    else:
        error_columns = ['drift_error_mean', 'linearity_error_mean', 'VSMOW_error_mean', 'replicate_dC_sem']

    stats['total_uncertainty'] = np.sqrt(np.sum([stats[col] ** 2 for col in error_columns], axis=0))

    return stats


from scipy.optimize import curve_fit
import numpy as np
import pandas as pd

def log_func(x, a, b, c):
    """
    a * exp(-b * x) + c
    a: amplitude (scales the height of the curve)
    b: decay rate
    c: offset (vertical shift)
    """
    return a * np.exp(b * x) + c

def exp_func(x, a, b, c):
    """
    Returns a * exp(b * x) + c
    """
    return a * np.exp(b * x) + c

def linear_func(x, a, b):
    """Linear function: f(x) = a*x + b"""
    return a * x + b

def guess_exponential_params(x, y):
    # x and y are 1D numpy arrays
    x = np.array(x)
    y = np.array(y)
    c0 = y.min()  # offset guess
    y_adjusted = np.array(y - c0)
    # To avoid log of zero or negative, ensure the first and last points are above c0
    # If they are not, you may need a more robust method
    if (y_adjusted <= 0).any():
        # Fallback: set offset to 0 if min(y) is negative or zero
        c0 = 0
        y_adjusted = np.array(y)
    a0 = y_adjusted[0]  # amplitude guess
    x = list(x)
    if len(x) > 1 and (y_adjusted[0] > 0) and (y_adjusted[-1] > 0):
        b0 = np.log(y_adjusted[-1] / y_adjusted[0]) / (x[-1] - x[0])
    else:
        b0 = 0.0  # fallback if we have too few points or invalid ratio

    return (a0, b0, c0)

def guess_log_params(x, y):
    x = np.array(x)
    y = np.array(y)
    # x must be > 0 for log
    b0 = 1.0
    # Handle case where x[0], x[-1] might be the same or cause log(0):
    if len(x) > 1 and x[0] > 0 and x[-1] > 0 and x[0] != x[-1]:
        a0 = (y[-1] - y[0]) / (np.log(x[-1]) - np.log(x[0]))
        c0 = y[0] - a0 * np.log(b0 * x[0])
    else:
        a0 = 1.0
        c0 = 0.0

    return (a0, b0, c0)


def guess_linear_params(x, y):
    """Estimate initial parameters for a linear fit using simple linear regression."""
    # Using numpy.polyfit to obtain slope and intercept
    slope, intercept = np.polyfit(x, y, 1)
    return (slope, intercept)


# def fit_and_select_best(x, y):
#     # Exponential fit
#     p0_exp = guess_exponential_params(x, y)
#     popt_exp, pcov_exp = curve_fit(exp_func, x, y, p0=p0_exp, maxfev=20000)
#     # Evaluate SSE for exponential
#     residuals_exp = y - exp_func(x, *popt_exp)
#     sse_exp = np.sum(residuals_exp**2)
#
#     # Log fit
#     p0_log = guess_log_params(x, y)
#     popt_log, pcov_log = curve_fit(log_func, x, y, p0=p0_log, maxfev=20000)
#     # Evaluate SSE for log
#     residuals_log = y - log_func(x, *popt_log)
#     sse_log = np.sum(residuals_log**2)
#
#     # Compare SSE (lower is better)
#     if sse_exp < sse_log:
#         return "exponential", popt_exp, sse_exp, pcov_exp
#     else:
#         return "log", popt_log, sse_log, pcov_log
def fit_and_select_best(x, y):
    # ---- Linear Fit ----
    p0_lin = guess_linear_params(x, y)
    popt_lin, pcov_lin = curve_fit(linear_func, x, y, p0=p0_lin, maxfev=20000)
    residuals_lin = y - linear_func(x, *popt_lin)
    sse_lin = np.sum(residuals_lin**2)

    # ---- Log Fit ----
    p0_log = guess_log_params(x, y)
    popt_log, pcov_log = curve_fit(log_func, x, y, p0=p0_log, maxfev=20000)
    residuals_log = y - log_func(x, *popt_log)
    sse_log = np.sum(residuals_log**2)

    # Compare SSE: lower is better.
    if sse_lin < sse_log:
        return "linear", popt_lin, sse_lin, pcov_lin
    else:
        return "log", popt_log, sse_log, pcov_log

# def prediction_std(best_model, x, popt, pcov):
#     """
#     Compute the standard error (1-sigma) of the predictions from exp_func
#     given parameters popt and covariance pcov.
#     """
#     if best_model == "exponential":
#         a, b, c = popt
#
#         # Jacobian partial derivatives at each x:
#         d_da = np.exp(b*x)       # ∂f/∂a
#         d_db = a * x * np.exp(b*x)  # ∂f/∂b
#         d_dc = np.ones_like(x)   # ∂f/∂c
#
#         # Stack into shape (n_points, 3)
#         # Each row: [d_da, d_db, d_dc]
#         J = np.vstack([d_da, d_db, d_dc]).T
#
#         # For each x_i, Var(f_i) = J_i * pcov * J_i^T
#         # We'll compute row by row:
#         var_pred = []
#         for row in J:
#             # row shape is (3,)
#             # reshape to (1, 3) for matrix multiplication
#             jac_row = row.reshape(1, -1)
#             var_i = jac_row @ pcov @ jac_row.T  # scalar
#             var_pred.append(var_i[0, 0])
#
#         return np.sqrt(var_pred)  # 1-sigma errors
#     else:
#
#         """
#         Computes the 1-sigma standard error of predictions for the model:
#             f(x) = a * ln(b*x) + c
#
#         Parameters
#         ----------
#         x : array-like
#             The x-values at which to compute prediction errors (must be > 0).
#         popt : array-like
#             Best-fit parameters [a, b, c] from curve_fit.
#         pcov : 2D array
#             Covariance matrix of the parameters from curve_fit.
#
#         Returns
#         -------
#         errors : np.ndarray
#             1D array of predicted 1-sigma standard errors for each x.
#
#         Notes
#         -----
#         The partial derivatives are:
#             ∂f/∂a = ln(b*x)
#             ∂f/∂b = a / b
#             ∂f/∂c = 1
#         so the Jacobian row for each x_i is [ln(b*x_i),  a/b,  1].
#         """
#         a, b, c = popt
#         x = np.array(x, dtype=float)
#
#         # Jacobian partial derivatives for each x
#         d_da = np.log(b * x)         # ∂f/∂a = ln(b*x)
#         d_db = np.full_like(x, a/b)  # ∂f/∂b = a / b (constant w.r.t x)
#         d_dc = np.ones_like(x)       # ∂f/∂c = 1
#
#         # Construct Jacobian: shape (n_points, 3)
#         J = np.vstack([d_da, d_db, d_dc]).T
#
#         var_pred = []
#         for row in J:
#             # row is shape (3,) => reshape to (1, 3) for matrix multiplication
#             jac_row = row.reshape(1, -1)
#             var_i = jac_row @ pcov @ jac_row.T  # scalar (1x1 matrix)
#             var_pred.append(var_i[0, 0])
#
#         return np.sqrt(var_pred)  # 1-sigma standard error
def prediction_std(best_model, x, popt, pcov):
    if best_model == "linear":
        # Linear model: f(x) = a*x + b
        # Partial derivatives: ∂f/∂a = x, ∂f/∂b = 1.
        x = np.array(x, dtype=float)
        d_da = x
        d_db = np.ones_like(x)
        J = np.vstack([d_da, d_db]).T  # shape (n_points, 2)

        var_pred = []
        for row in J:
            jac_row = row.reshape(1, -1)  # shape (1,2)
            var_i = jac_row @ pcov @ jac_row.T
            var_pred.append(var_i[0, 0])
        return np.sqrt(var_pred)  # 1-sigma errors

    elif best_model == "log":
        # Log fit: f(x) = a * ln(b*x) + c
        a, b, c = popt
        x = np.array(x, dtype=float)
        d_da = np.log(b * x)         # ∂f/∂a
        d_db = np.full_like(x, a/b)    # ∂f/∂b (constant with respect to x)
        d_dc = np.ones_like(x)         # ∂f/∂c
        J = np.vstack([d_da, d_db, d_dc]).T  # shape (n_points, 3)

        var_pred = []
        for row in J:
            jac_row = row.reshape(1, -1)
            var_i = jac_row @ pcov @ jac_row.T
            var_pred.append(var_i[0, 0])
        return np.sqrt(var_pred)

    else:
        raise ValueError("Unknown model type for prediction_std.")


import os
import datetime 
from .define_standards import open_editor 

def append_to_log(log_file_path, log_message):
    """
    Add entry to log file.
    """
    with open(log_file_path, "a") as log_file:
        current_datetime = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        initial_message = f"Log file created at {current_datetime}\n"
        log_file.write(log_message + "; " + str(current_datetime) + "\n")


def pos_response(response):
    return response.lower() in {"yes", "y", "true", "t", ""}


def neg_response(response):
    return response.lower() in {"no", "n", "false", "f"}


def query_project_name():
    project_name = input("\nProvide the project name.\n")
    return project_name 


def query_file_location():
    while True:
        loc = input("\nProvide the full path of the GC-IRMS datafile (as .csv).\n")
        
        # Remove single quotes if present at both the start and end of the input
        if loc.startswith("'") and loc.endswith("'"):
            loc = loc[1:-1]
        
        if os.path.isfile(loc) and loc.endswith(".csv"):
            return loc
        else:
            print("\nFile does not exist or is not a .csv file. Try again.\n")

def isotope_type():
    """
    Ask the user for their choice isotope.
    """
    while True:  # Start an infinite loop
        choice = input("\nCorrecting carbon or hydrogen? (c/h)\n")

        if choice in ["c", "h"]:
            if choice == "c":
                isotope = "dC"
            else:
                isotope = "dD"
            return isotope
        else:
            print("\nInvalid response\n")

def query_stds(alt_stds, isotope):
    standards_df = open_editor(alt_stds, isotope)
    return standards_df


def lin_response(log_file_path):
    valid_responses = ["yes", "y", "true", "t", "no", "n", "false", "f"]
    while True:
        response = input("\nAssign a linearity correction? (Y/N)\n").lower()
        if response in valid_responses:
            append_to_log(log_file_path, "Linearity application application: " + str(response))
            return response
        else:
            print("\nInvalid response. Try again.\n")


def q_methylation(unknown, stds, log_file_path):  # , user_choice, response):
    from .corrections.methanol import methyl_correction
    while True:
        response = input("\nMethanol δD is -72.5 ± 3.1 ‰. Is this correct? (Y/N)\n").lower()
        if pos_response(response):
            meth_dD = -72.5
            meth_std = 3.1
            unknown = methyl_correction(unknown, stds)
            break
        elif neg_response(response):
            meth_dD = input("\nMeasured δD value of the methanol used in FAME methylation?\n")
            meth_std = input("\nUncertainty of the methanol δD value?\n")
            unknown, stds = methyl_correction(unknown, stds, mdD=meth_dD, mdD_err=meth_std)
            break
        else:
            print("\nInvalid response. Try again.\n")
    append_to_log(log_file_path, f"Methanol δD: {meth_dD} ± {meth_std} ‰")
    return unknown, stds

def q_original_phthalic_value():
    o_ph_dD = input("Enter isotopic value of phthalic acid: ")
    return o_ph_dD

def q_output():
    o_fp = input("Provide a folder path for the output data:\n")
    return o_fp


import pandas as pd

def methyl_correction(unknown, stds, mdD = -72.5, mdD_err = 3.1):
    """
    Correct FAMES for δD of methyl groups introduced 
    during methylation.
    ~GAO~ 12/4/2023
    """
    # Extract the number of carbons from the 'chain' column
    #c_n = unknown.loc[unknown['chain']!="Phthalic acid", 'chain'].str.extract('C(\d+)').astype(int).squeeze()
    c_n = unknown.loc[unknown['chain']!="Phthalic acid", 'chain'].str.extract(r'C(\d+)').astype(int).squeeze()
    # Apply the correction formula
    unknown.loc[unknown['chain']!="Phthalic acid", 'methanol_dD'] = ((unknown['VSMOW_dD'] * (2 * c_n + 2)) - (mdD * 3)) / (2 * c_n)
    unknown.loc[unknown['chain']!="Phthalic acid",'methanol_error'] = mdD_err
    return unknown
